/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.lib.path;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.function.Consumer;

import com.google.common.base.Throwables;
import com.google.common.util.concurrent.Futures;

import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;

import buildcraft.lib.misc.WorkerThreadUtil;
import buildcraft.lib.path.task.TaskMiniChunkManager;

public class MiniChunkCache {
    private static Map<Integer, MiniChunkCache> worldCaches = new HashMap<>();

    public final int dimId;
    private final Map<BlockPos, MiniChunkGraph> cache = new ConcurrentHashMap<>();
    final Map<BlockPos, Future<MiniChunkGraph>> tempData = new ConcurrentHashMap<>();

    private MiniChunkCache(int dimId) {
        this.dimId = dimId;
    }

    public static Future<MiniChunkGraph> requestGraph(World world, BlockPos pos) {
        int dimId = world.field_73011_w.getDimension();
        if (!worldCaches.containsKey(dimId)) {
            worldCaches.put(dimId, new MiniChunkCache(dimId));
        }
        return worldCaches.get(dimId).requestGraphImpl(world, pos);
    }

    public static MiniChunkGraph getGraphIfExists(World world, BlockPos pos) {
        int dimId = world.field_73011_w.getDimension();
        if (!worldCaches.containsKey(dimId)) {
            worldCaches.put(dimId, new MiniChunkCache(dimId));
        }
        return worldCaches.get(dimId).getGraphIfExistsImpl(pos);
    }

    public static MiniChunkGraph requestAndWait(World world, BlockPos pos) {
        try {
            return requestGraph(world, pos).get();
        } catch (InterruptedException | ExecutionException e) {
            throw Throwables.propagate(e);
        }
    }

    private void putGraph(BlockPos min, MiniChunkGraph graph) {
        cache.put(min, graph);
    }

    private Future<MiniChunkGraph> requestGraphImpl(World world, BlockPos pos) {
        final BlockPos minPos = convertToMin(pos);
        pos = minPos;
        MiniChunkGraph existing = cache.get(pos);
        if (existing != null) {
            return Futures.immediateCheckedFuture(existing);
        }
        if (!world.func_175667_e(pos)) return Futures.immediateFailedFuture(new Throwable("The block " + pos + " is not loaded!"));
        synchronized (this) {
            if (tempData.containsKey(pos)) {
                return tempData.get(pos);
            }
            Consumer<MiniChunkGraph> setter = (graph) -> putGraph(minPos, graph);
            Callable<MiniChunkGraph> task = new TaskMiniChunkManager(world, pos, setter);
            Future<MiniChunkGraph> future = WorkerThreadUtil.executeDependantTask(task);
            tempData.put(pos, future);
            return future;
        }
    }

    private MiniChunkGraph getGraphIfExistsImpl(BlockPos pos) {
        pos = convertToMin(pos);
        return cache.get(pos);
    }

    private static BlockPos convertToMin(BlockPos pos) {
        // Get the minimum blockpos of the minichunk
        int x = (pos.func_177958_n() / 16) * 16;
        int y = (pos.func_177956_o() / 16) * 16;
        int z = (pos.func_177952_p() / 16) * 16;
        return new BlockPos(x, y, z);
    }
}
